#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/vec.h"
#include "prjlibs-c/check_close.h"
#include "types.h"
#include "constants.h"
#include "fmt0.h"
#include "must.h"
#include "get_unique_id.h"
#include "dirsync.h"

ccp PROG="posterity-deliver";

static str path_tmp=VEC_0(char), path_new=VEC_0(char);
static buffer out;
static char bufspace[BUFFER_OUTSIZE];
static ccp rpline, dtline;

static type_status get(char* ch) {
  if (*rpline!='\0') { *ch=*rpline++; return 1; }
  if (*dtline!='\0') { *ch=*dtline++; return 1; }
  return buffer_get(buffer_0, ch, 1);
}

static void must_put(char ch) {
  if (buffer_put(&out, &ch, 1)<0) DIE1(write, s(&path_tmp));
}

int main(int argc, char** argv) {
  struct stat statbuf;
  str unique_id=VEC_0(char);
  size_t path_len_new;
  type_fd fd;
  mode_t mask;
  ccp convert;

  if (argc<2 || argv[1][0]=='\0') DIE_USAGE(" maildir [command...]");

  if (stat(argv[1], &statbuf)!=0) DIE1(stat, argv[1]);
  if (S_ISVTX&statbuf.st_mode) _exit(111);

  rpline =getenv(env_rpline ); if (rpline ==NULLP) rpline ="";
  dtline =getenv(env_dtline ); if (dtline ==NULLP) dtline ="";
  convert=getenv(env_convert); if (convert==NULLP) convert="1";
  if (strcmp(convert, "0")==0) convert=NULLP;

  get_unique_id(&unique_id);
  must_s(&path_tmp, argv[1], 1);
  must_s(&path_new, argv[1], 1);
  if (s(&path_tmp)[l(&path_tmp)-1]!='/') {
    must_1(&path_tmp, '/');
    must_1(&path_new, '/');
  }
  must_a(&path_tmp, subdir_tmp, 1);
  must_a(&path_new, subdir_new, 1);
  path_len_new=l(&path_new)-1;
  must_n(&path_tmp, s(&unique_id), l(&unique_id));
  must_n(&path_new, s(&unique_id), l(&unique_id));
  must_1(&path_tmp, '\0');
  must_1(&path_new, '\0');
  if (pathexec_env(env_message, s(&path_new))==0) DIE0(alloc);

  mask=umask(0);
  fd=open(s(&path_tmp), O_WRONLY|O_CREAT|O_EXCL, 0444|(0222&~mask));
  if (fd<0) DIE1(open, s(&path_tmp));
  umask(mask);
  buffer_init(&out, &fd_writev, fd, bufspace, sizeof bufspace);

  for (;;) {
    char ch;
    type_status status=get(&ch);
    if (status==0) break;
    if (status<0) DIE1(read, src_stdin);
    if (ch=='\n' && convert!=NULLP) must_put('\r');
    must_put(ch);
  }

  if (buffer_flush(&out)==0) DIE1(write, s(&path_tmp));
  if (fsync(fd)!=0)          DIE1(fsync, s(&path_tmp));
  if (check_close(fd)!=0)    DIE1(close, s(&path_tmp));

  if (link(s(&path_tmp), s(&path_new))!=0)
    DIE2(link, s(&path_tmp), s(&path_new));
  if (unlink(s(&path_tmp))!=0) WARN1(unlink, s(&path_tmp));

  *VEC_ELT(char, &path_new, path_len_new)='\0';
  if (dirsync(s(&path_new))!=0) DIE1(write, s(&path_new));

  argv+=2;
  pathexec0((ccp*)argv);
  DIE1(exec, argv[0]);
}
