#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/vec.h"
#include "prjlibs-c/check_close.h"
#include "fmt0.h"
#include "types.h"

ccp PROG="posterity-deliver";

static VEC_T(char) path_tmp=VEC_0(char), path_new=VEC_0(char);
static buffer out;
static char bufspace[BUFFER_OUTSIZE];
static char const* rpline, * dtline;

static type_status get(char* ch) {
  if (*rpline!='\0') { *ch=*rpline++; return 1; }
  if (*dtline!='\0') { *ch=*dtline++; return 1; }
  return buffer_get(buffer_0, ch, 1);
}

static void must_put(char ch) {
  if (buffer_put(&out, &ch, 1)<0) DIE1(write, VEC_ELTS(char, &path_tmp));
}

int main(int argc, char** argv) {
  struct stat statbuf;
  struct timeval tv;
  type_uint64_buf buf;
  type_len len;
  stralloc hostname=GENALLOC_ZERO;
  size_t path_len_new;
  type_fd fd;
  mode_t mode;

  if (argc<2 || argv[1][0]=='\0') DIE_USAGE(" maildir [command...]");

  if (stat(argv[1], &statbuf)!=0) DIE1(stat, argv[1]);
  if (S_ISVTX&statbuf.st_mode) _exit(111);

  rpline=getenv("RPLINE"); if (rpline==NULLP) rpline="";
  dtline=getenv("DTLINE"); if (dtline==NULLP) dtline="";

  gettimeofday(&tv, NULLP);
  if (sagethostname(&hostname)!=0) DIE0(gethostname);

  if (VEC_PUSHS(&path_tmp, argv[1], 1)!=0 ||
      VEC_PUSHS(&path_new, argv[1], 1)!=0 ||
      VEC_PUSHA(char, &path_tmp, "/tmp/", 1)!=0 ||
      VEC_PUSHA(char, &path_new, "/new/", 1)!=0 ||
      (path_len_new=VEC_LEN(char, &path_new)-1, 0) ||
      VEC_PUSHN(char, &path_tmp, buf, len=uint64_fmt(buf, tv.tv_sec))!=0 ||
      VEC_PUSHN(char, &path_new, buf, len)!=0 ||
      VEC_PUSHA(char, &path_tmp, ".M", 1)!=0 ||
      VEC_PUSHA(char, &path_new, ".M", 1)!=0 ||
      VEC_PUSHN(char, &path_tmp, buf, len=uint64_fmt(buf, tv.tv_usec))!=0 ||
      VEC_PUSHN(char, &path_new, buf, len)!=0 ||
      VEC_PUSH1(char, &path_tmp, 'P')!=0 ||
      VEC_PUSH1(char, &path_new, 'P')!=0 ||
      VEC_PUSHN(char, &path_tmp, buf, len=uint64_fmt(buf, getpid()))!=0 ||
      VEC_PUSHN(char, &path_new, buf, len)!=0 ||
      VEC_PUSH1(char, &path_tmp, '.')!=0 ||
      VEC_PUSH1(char, &path_new, '.')!=0 ||
      VEC_PUSHN(char, &path_tmp, hostname.s, hostname.len)!=0 ||
      VEC_PUSHN(char, &path_new, hostname.s, hostname.len)!=0 ||
      VEC_PUSH1(char, &path_tmp, '\0')!=0 ||
      VEC_PUSH1(char, &path_new, '\0')!=0)
    DIE0(alloc);

  mode=umask(0);
  umask(mode & 0333);
  fd=open(VEC_ELTS(char, &path_tmp), O_WRONLY|O_CREAT|O_EXCL, 0666);
  if (fd<0) DIE1(open, VEC_ELTS(char, &path_tmp));
  umask(mode);
  buffer_init(&out, &fd_writev, fd, bufspace, sizeof bufspace);

  for (;;) {
    char ch;
    type_status status=get(&ch);
    if (status==0) break;
    if (status<0) DIE1(read, "standard input");
    if (ch=='\n') must_put('\r');
    must_put(ch);
  }

  if (buffer_flush(&out)==0) DIE1(write, VEC_ELTS(char, &path_tmp));
  if (fsync(fd)!=0)          DIE1(fsync, VEC_ELTS(char, &path_tmp));
  if (check_close(fd)!=0)    DIE1(close, VEC_ELTS(char, &path_tmp));

  if (link(VEC_ELTS(char, &path_tmp), VEC_ELTS(char, &path_new))!=0)
    DIE2(link, VEC_ELTS(char, &path_tmp), VEC_ELTS(char, &path_new));
  unlink(VEC_ELTS(char, &path_tmp));

  *VEC_ELT(char, &path_new, path_len_new)='\0';
  fd=open(VEC_ELTS(char, &path_new), O_RDONLY);
  if (fd<0)               DIE1(open,  VEC_ELTS(char, &path_new));
  if (fsync(fd)!=0)       DIE1(fsync, VEC_ELTS(char, &path_new));
  if (check_close(fd)!=0) DIE1(close, VEC_ELTS(char, &path_new));

  argv+=2;
  exec0((char const**)argv);
  DIE1(exec, argv[0]);
}
