#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <skalibs/stddjb.h>
#include <skalibs/cdb_make.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/intattr.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/vec.h"
#include "prjlibs-c/nsec.h"
#include "types.h"
#include "constants.h"
#include "must.h"
#include "get_unique_id.h"
#include "dirsync.h"
#include "keys.h"
#include "b0.h" 
#include "must.h" 
#include "fmt0.h" 

ccp PROG="posterity-update";

#define DIE_EOH() DIE2X(100, "unterminated message header", "")

static type_uid_buf const empty_uid_buf="";

static str unique_id=VEC_0(char);
static str path_msg=VEC_0(char);
static str path_tmpinfo=VEC_0(char);
static str path_msginfo=VEC_0(char);
static str path_uidinfo=VEC_0(char);
static str path_uid_open=VEC_0(char);
static str path_uid_link=VEC_0(char);
static size_t path_msg_len;
static size_t path_msginfo_len;
static size_t path_uidinfo_len;
static size_t path_uid_len;

typedef struct {
  str uniq;
  type_uid uid;
  time_t sec;
  unsigned int usec;
} type_msg;
VEC_TD(type_msg, type_msgvec);

static void maybe_mkdir(ccp path) {
  struct stat statbuf;
  if (stat(path, &statbuf)==0 && S_ISDIR(statbuf.st_mode)) return;
  if (mkdir(path, 0777)!=0 && errno!=EEXIST) DIE1(mkdir, path);
}

static void setup(void) {
  ccp slash;
  size_t path_maildir_len;
  ccp const msg=getenv(env_message);
  if (msg==NULLP) DIE3X(100, "$", env_message, " must be set");
  slash=strrchr(msg, '/');
  if (slash==NULLP || slash-msg<4 || memcmp(slash-4, "/new", 4)!=0)
    DIE3X(100, "$", env_message,
          " must be a message in a maildir new/ directory");
  path_maildir_len=slash-msg-3;
  must_n(&path_msg,      msg, path_maildir_len);
  must_n(&path_uid_open, msg, path_maildir_len);
  must_n(&path_uid_link, msg, path_maildir_len);
  must_n(&path_msginfo,  msg, path_maildir_len);
  must_n(&path_tmpinfo,  msg, path_maildir_len);
  must_n(&path_uidinfo,  msg, path_maildir_len);
  must_a(&path_msg,      subdir_new,       0);
  must_a(&path_uid_open, subdir_uidlock,   0);
  must_a(&path_uid_link, subdir_uidlock,   0);
  must_a(&path_msginfo,  subdir_msginfo,   0);
  must_a(&path_tmpinfo,  subdir_msginfo,   0);
  must_a(&path_uidinfo,  subdir_posterity, 0);
  path_msg_len    =l(&path_msg);
  path_uid_len    =l(&path_uid_open)-1;
  path_msginfo_len=l(&path_msginfo)-1;
  maybe_mkdir(s(&path_uidinfo));
  VEC_TRUNC(char, &path_uidinfo, path_maildir_len);
  must_a(&path_uidinfo, subdir_uidinfo, 0);
  path_uidinfo_len=l(&path_uidinfo)-1;
  maybe_mkdir(s(&path_uid_open));
  maybe_mkdir(s(&path_msginfo));
  maybe_mkdir(s(&path_uidinfo));
  get_unique_id(&unique_id);
}

static void get_messages(type_msgvec* msgs) {
  DIR* dirh;
  size_t i, n_msgs=VEC_LEN(type_msg, msgs);
  VEC_TRUNC(char, &path_msg, path_msg_len);
  must_1(&path_msg, '\0');
  dirh=opendir(s(&path_msg));
  if (dirh==NULLP) DIE1(open, s(&path_msg));
  for (;;) {
    type_msg* msg;
    size_t uniq_len;
    struct dirent* de=(errno=0, readdir(dirh));
    if (de==NULLP) {
      if (errno==0) break;
      DIE1(read, s(&path_msg));
    }
    if (de->d_name[0]=='.') continue;
    uniq_len=strlen(de->d_name)+1;
    if (uniq_len==0) DIE_OVERFLOW();
    /* if (kart_insert(uniq)==already_exists) continue; */ 
    for (i=0; i!=n_msgs; ++i) {
      str* uniq=&VEC_ELT(type_msg, msgs, i)->uniq;
      if (uniq_len==l(uniq) && memcmp(de->d_name, s(uniq), uniq_len-1)==0)
        break;
    }
    if (i!=VEC_LEN(type_msg, msgs)) continue;
    /* end kart */ 
    if (VEC_PUSHN(type_msg, msgs, NULLP, 1)!=0) DIE0(alloc);
    msg=VEC_ELT(type_msg, msgs, VEC_LEN(type_msg, msgs)-1);
    VEC_INIT(char, &msg->uniq);
    must_s(&msg->uniq, de->d_name, 0);
  }
  closedir(dirh);
  for (i=0; i!=VEC_LEN(type_msg, msgs); ++i) {
    struct stat statbuf;
    type_msg* msg=VEC_ELTS(type_msg, msgs)+i;
    if (msg->usec!=0) continue;
    VEC_TRUNC(char, &path_msg, path_msg_len);
    must_n(&path_msg, s(&msg->uniq), l(&msg->uniq));
    if (stat(s(&path_msg), &statbuf)!=0) DIE1(stat, s(&path_msg));
    msg->sec=statbuf.st_mtime;
    msg->usec=1+nsec_mtime(&statbuf);
  }
}

static type_status maybe_mkfile(ccp path) {
  type_fd const fd=open3(path, O_WRONLY|O_CREAT|O_EXCL, 0600);
  if (fd<0) {
    if (errno==EEXIST) return -1;
    DIE1(open, path);
  }
  if (fd_close(fd)!=0) DIE1(close, path);
  return 0;
}

static type_uid allocate_uid(void) {
  type_uid uid_open, uid_link=0, last_uid_link, nlink;
  type_bool flag_try_open;
  char* path_open, * path_link;
  VEC_TRUNC(char, &path_uid_open, path_uid_len);
  VEC_TRUNC(char, &path_uid_link, path_uid_len);
  must_a(&path_uid_open, empty_uid_buf, 0);
  must_a(&path_uid_link, empty_uid_buf, 0);
  path_open=s(&path_uid_open);
  path_link=s(&path_uid_link);
  for (flag_try_open=1, uid_open=1;;) {
    struct stat statbuf;
    ino_t ino_open;
    if (flag_try_open) {
      fmt_uid(path_open+path_uid_len, uid_open);
      if (maybe_mkfile(path_open)==0) return uid_open;
      flag_try_open=0;
      last_uid_link=0;
    } else
      last_uid_link=uid_link;
    if (stat(path_open, &statbuf)!=0) DIE1(stat, path_open);
    ino_open=statbuf.st_ino;
    nlink=statbuf.st_nlink;
    uid_link=uid_open+nlink;
    if (INTATTR_NEGATIVE(nlink_t, statbuf.st_nlink) ||
        (nlink_t)nlink!=statbuf.st_nlink || uid_link<nlink)
      DIE_OVERFLOW();
    if (uid_link==last_uid_link)
      /* We already tried this and got EEXIST; this file's links aren't
         numbered consecutively. */
      DIE2X(100, "corrupt uidlock: ", path_open);
    fmt_uid(path_link+path_uid_len, uid_link);
    if (link(path_open, path_link)==0) return uid_link;
    switch (errno) {
      case EMLINK: /* Too many hard links; start a new file. */
        flag_try_open=1;
        uid_open=uid_link;
        break;
      case EEXIST:
        if (stat(path_link, &statbuf)!=0) DIE1(stat, path_link);
        if (statbuf.st_ino!=ino_open) {
          str_copy(path_open+path_uid_len, path_link+path_uid_len);
          uid_open=uid_link;
          uid_link=0;
        }
        /* else we probably just lost a race */ 
        break;
      default: DIE2(link, path_open, path_link);
    }
  }
}

static void load_uid(type_msg* msg) {
  str* const uniq=&msg->uniq;
  struct cdb c;
  type_fd fd_info;
  type_len uid_len;
  type_uid_buf uid_buf="";
  VEC_TRUNC(char, &path_msginfo, path_msginfo_len);
  must_n(&path_msginfo, s(uniq), l(uniq));
  fd_info=open2(s(&path_msginfo), O_RDONLY);
  if (fd_info<0) DIE1(open, s(&path_msginfo));
  if (cdb_init(&c, fd_info)!=0) DIE1(read, s(&path_msginfo));
  cdb_findstart(&c);
  switch (cdb_findnext(&c, k(uid))) {
    case 0: DIE2X(100, "corrupt cdb: ", s(&path_msginfo));
    case -1: DIE1(read, s(&path_msginfo));
  }
  uid_len=cdb_datalen(&c);
  if (uid_len==0 || uid_len>=sizeof uid_buf)
    DIE2X(100, "corrupt cdb: ", s(&path_msginfo));
  if (cdb_read(&c, uid_buf, uid_len, cdb_datapos(&c))!=0)
    DIE1(read, s(&path_msginfo));
  cdb_free(&c);
  fd_close(fd_info);
  if (scan_uid(uid_buf, &msg->uid)!=uid_len)
    DIE2X(100, "corrupt cdb: ", s(&path_msginfo));
}

/*
static void must_put(buffer* buf, ccp dest, ccp data, type_len len,
                     type_fsize* total) {
  if (buffer_put(buf, data, len)<0) DIE1(write, dest);
  if (total==NULLP) return;
  *total+=len;
  if (*total<len) DIE_OVERFLOW();
}
*/

static void must_put_cdb(struct cdb_make* cm, ccp dst, ccp key, uint32 klen,
                         ccp data, uint32 dlen) {
  if (cdb_make_add(cm, key, klen, data, dlen)!=0) DIE1(write, dst);
}

/*
static type_len must_fill_line(buffer* buf, ccp src) {
  for (;;) {
    type_len const eo_line=byte_chr(buffer_RPEEK(buf), buffer_len(buf), '\n');
    if (eo_line!=buffer_len(buf)) return eo_line;
    if (must_fill(buf, src)==0) DIE_EOH();
  }
}
*/

typedef struct mimepart mimepart;
struct mimepart {
  type_fsize mime0, mime1, head0, head1, text0, text1;
  mimepart* parent;
  unsigned int partnum;
  type_bool flag_multipart:1, flag_multipart_digest:1, flag_mess822:1;
  char boundary[1];
};
#define MIMEPART_SIZE(boundlen) (offsetof(mimepart, boundary)+boundlen)

static void parse_message(type_msg* msg) {
  str* const uniq=&msg->uniq;
  type_fd fd_msg, fd_info;
  struct cdb_make cm;
  //type_len buflen; 
  //type_bool flag_header;
  //type_uint64_buf buf_64;
  char buf_cdb[2*sizeof (type_uint64_buf)];
  //type_fsize total=0;
  VEC_TRUNC(char, &path_msginfo, path_msginfo_len);
  must_n(&path_msginfo, s(uniq), l(uniq));
  { /* If this message has already been parsed, bail out. */
    struct stat statbuf;
    if (stat(s(&path_msginfo), &statbuf)==0) {
      load_uid(msg);
      return;
    }
    if (errno!=ENOENT) DIE1(stat, s(&path_msginfo));
  }
  VEC_TRUNC(char, &path_tmpinfo, path_msginfo_len);
  must_1(&path_tmpinfo, '.');
  must_n(&path_tmpinfo, s(&unique_id), l(&unique_id));
  must_1(&path_tmpinfo, '\0');
  VEC_TRUNC(char, &path_msg, path_msg_len);
  must_n(&path_msg, s(uniq), l(uniq));

  fd_info=open3(s(&path_tmpinfo), O_WRONLY|O_CREAT|O_EXCL, 0666);
  if (fd_info<0) DIE1(open, s(&path_tmpinfo));
  if (cdb_make_start(&cm, fd_info)!=0) DIE1(write, s(&path_tmpinfo));
  must_put_cdb(&cm, s(&path_tmpinfo), k(uniq), sl(uniq));

  fd_msg=open2(s(&path_msg), O_RDONLY);
  if (fd_msg<0) DIE1(open, s(&path_msg));

/*
  for (flag_header=1;;) {
    type_len eo_line, field_offset;
    type_fsize eo_field, value_start;
    type_bool flag_mime;
    if (!flag_header) {
      if (buffer_len(buffer_0)==0)
        if (must_fill(buffer_0, src_stdin)==0) break;
      eo_line=byte_chr(buffer_RPEEK(buffer_0), buffer_len(buffer_0), '\n');
      must_put(&msgbuf, path_msg_tmp.s, buffer_RPEEK(buffer_0), eo_line,
               &total);
      if (eo_line!=buffer_len(buffer_0)) {
        must_put(&msgbuf, path_msg_tmp.s, "\r\n", 2, &total);
        ++eo_line;
      }
      buffer_RSEEK(buffer_0, eo_line);
      continue;
    }

    eo_line=must_fill_line(buffer_0, src_stdin);
    if (eo_line==0) {
      flag_header=0;
      must_put_cdb(&cm, s(&path_tmpinfo), k(body), buf_cdb,
                   fmt_fsize(buf_cdb, total+2));
      continue;
    }

    eo_field=byte_chr(buffer_RPEEK(buffer_0), eo_line, ':');
    if (eo_field==eo_line) DIE_EOH();
    MUST_COPYS(&key, "H:"); /cdb header field key/ 
    field_offset=key.len;
    must_catb(&key, buffer_RPEEK(buffer_0), eo_field);
    case_lowerb(key.s+field_offset, eo_field);
    ++eo_field;
    eo_field+=total;
    value_start=total+eo_field;

    must_put(&msgbuf, path_msg_tmp.s, buffer_RPEEK(buffer_0), eo_line, &total);
    must_put(&msgbuf, path_msg_tmp.s, "\r\n", 2, &total);
    buffer_RSEEK(buffer_0, eo_line+1);

    if (buffer_len(buffer_0)==0) {
      if (must_fill(buffer_0, src_stdin)==0) DIE_EOH();
    }
    while (buffer_RPEEK(buffer_0)[0]==' ' ||
           buffer_RPEEK(buffer_0)[0]=='\t') {
      eo_line=must_fill_line(buffer_0, src_stdin);
      must_put(&msgbuf, path_msg_tmp.s, buffer_RPEEK(buffer_0), eo_line,
               &total);
      must_put(&msgbuf, path_msg_tmp.s, "\r\n", 2, &total);
      buffer_RSEEK(buffer_0, eo_line+1);
    }

    buflen=fmt_fsize(buf_cdb, eo_field);
    buf_cdb[buflen++]=':';
    buflen+=fmt_fsize(buf_cdb+buflen, total-2);
    must_put_cdb(&cm, s(&path_tmpinfo), key.s, key.len, buf_cdb, buflen);
  }
*/

  fd_close(fd_msg);

  /* We're about to burn a UID.  To minimize the chances of letting it go to
     waste, let's detect any errors we can beforehand, and check again
     whether another process already parsed this message. */
  if (fsync(fd_info)!=0) DIE1(write, s(&path_tmpinfo));
  { struct stat statbuf;
    if (stat(s(&path_msginfo), &statbuf)==0) {
      load_uid(msg);
      if (unlink(s(&path_tmpinfo))!=0) WARN1(unlink, s(&path_tmpinfo));
      return;
    }
    if (errno!=ENOENT) DIE1(stat, s(&path_msginfo));
  }

  msg->uid=allocate_uid();
  must_put_cdb(&cm, s(&path_tmpinfo), k(uid), buf_cdb,
               fmt_uid(buf_cdb, msg->uid));

  if (cdb_make_finish(&cm)<0 || fsync(fd_info)!=0 || fd_close(fd_info)!=0)
    DIE1(write, s(&path_tmpinfo));
  if (link(s(&path_tmpinfo), s(&path_msginfo))!=0) {
    if (errno!=EEXIST) DIE2(link, s(&path_tmpinfo), s(&path_msginfo));
    load_uid(msg);
  }
  VEC_TRUNC(char, &path_msginfo, path_msginfo_len);
  must_1(&path_msginfo, '\0');
  dirsync(s(&path_msginfo));
  if (unlink(s(&path_tmpinfo))!=0) WARN1(unlink, s(&path_tmpinfo));
}

static type_bool linked_any_uids=0;

static void link_uid(type_msg* msg) {
  str* const uniq=&msg->uniq;
  VEC_TRUNC(char, &path_msginfo, path_msginfo_len);
  VEC_TRUNC(char, &path_uidinfo, path_uidinfo_len);
  must_n(&path_msginfo, s(uniq), l(uniq));
  must_a(&path_uidinfo, empty_uid_buf, 0);
  fmt_uid(s(&path_uidinfo)+path_uidinfo_len, msg->uid);
  if (link(s(&path_msginfo), s(&path_uidinfo))!=0) {
    if (errno!=EEXIST) DIE2(link, s(&path_msginfo), s(&path_uidinfo));
  } else
    linked_any_uids=1;
}

static void notify_fifo(void) {
  
}

static void cleanup(void) {
  
}

static int sorter_time(void const* va, void const* vb) {
  type_msg const* a=va, * b=vb;
  if (a->sec <b->sec ) return -1;
  if (a->sec >b->sec ) return  1;
  if (a->usec<b->usec) return -1;
  if (a->usec>b->usec) return  1;
  return strcmp(s(&a->uniq), s(&b->uniq));
}

static int sorter_uid(void const* va, void const* vb) {
  type_msg const* a=va, * b=vb;
  if (a->uid<b->uid) return -1;
  if (a->uid>b->uid) return  1;
  return strcmp(s(&a->uniq), s(&b->uniq));
}

int main(int argc, char** argv) {
  type_msgvec msgs=VEC_0(type_msg);
  size_t i;
  (void)argc;
  setup();
  for (;;) {
    size_t n_msgs=VEC_LEN(type_msg, &msgs);
    get_messages(&msgs);
    if (n_msgs==VEC_LEN(type_msg, &msgs)) break;
    qsort(VEC_ELTS(type_msg, &msgs)+n_msgs, VEC_LEN(type_msg, &msgs)-n_msgs,
          sizeof (type_msg), sorter_time);
    for (i=n_msgs; i!=VEC_LEN(type_msg, &msgs); ++i)
      parse_message(VEC_ELTS(type_msg, &msgs)+i);
  }
  qsort(VEC_ELTS(type_msg, &msgs), VEC_LEN(type_msg, &msgs),
        sizeof (type_msg), sorter_uid);
  for (i=0; i!=VEC_LEN(type_msg, &msgs); ++i)
    link_uid(VEC_ELTS(type_msg, &msgs)+i);
  if (linked_any_uids) notify_fifo();
  cleanup();
  argv+=1;
  exec0((char const**)argv);
  DIE1(exec, argv[0]);
}
