#include "prjlibs-c/standards.h"
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/diewarn.h"
#include "timeout.h"

static int flag_timeout, flag_timelimit;
static long timeout;
static time_t timelimit;

void timeout_setup(void) {
  uint64_t l;
  unsigned int i;
  char const* var;
  var=getenv("TIMEOUT");
  flag_timeout=(var!=NULLP);
  if (flag_timeout) {
    i=uint64_scan(var, &l);
    if (i==0 || var[i]!='\0') DIE_MALFORMED("$TIMEOUT", var+i);
    timeout=l;
    if (timeout<(long)0 || l!=(unsigned long)timeout)
      DIE_OVERFLOW();
  }
  var=getenv("TIMELIMIT");
  flag_timelimit=(var!=NULLP);
  if (flag_timelimit) {
    time_t tmp;
    i=uint64_scan(var, &l);
    if (i==0 || var[i]!='\0') DIE_MALFORMED("$TIMELIMIT", var+i);
    timelimit=time(NULLP);
    tmp=timelimit+(time_t)l;
    if (tmp<timelimit || (unsigned long)(tmp-timelimit)!=l)
      DIE_OVERFLOW();
    timelimit=tmp;
  }
  timeout_reset();
}

void timeout_reset(void) {
  time_t t;
  struct itimerval timer;
  if (!flag_timeout && !flag_timelimit) return;
  t=time(NULLP);
  if (flag_timelimit) {
    if (timelimit>t) timer.it_value.tv_sec=timelimit-t;
    else {
      raise(SIGALRM);
      flag_timelimit=0;
      if (!flag_timeout) return;
    }
  }
  if (flag_timeout && (!flag_timelimit || timer.it_value.tv_sec>timeout))
    timer.it_value.tv_sec=timeout;
  timer.it_value.tv_usec=0;
  timer.it_interval.tv_sec=0;
  timer.it_interval.tv_usec=0;
  if (setitimer(ITIMER_REAL, &timer, NULLP)!=0) DIE0(setitimer);
}
